
# THIS FILE IS GENERATED FROM TINT SETUP.PY
short_version = '0.1.0'
version = '0.1.0'
full_version = '0.1.0.dev+d933665'
git_revision = 'd933665f514dfd429cf3691f87d500b370290873'
release = False

if not release:
    version = full_version
